require("LuaLib")
require("MonsterLib")

require("boids")
require("debug_ext")
require("math_ext")
require("util")
require("movement")
require("vec2")

-- Declare these at the global scope of this file, so we can access them
-- anywhere
local AIFlyingHash = hash("AIFlying")

local function dist_origin()
    return dist_point(param.origin.x, param.origin.y)
end

local function dist_player()
    local me, player = getMe(), getPlayer()
    -- 
    local vecMe = { x = me.XPos, y = me.YPos }
    local vecPlayer = { x = player.XPos, y = player.YPos }

    return vec_dist( vecMe, vecPlayer )
end

local function attack_swipe(data)
    if data ~= nil then
        data.dir = dirs.N
        data.pressAction = false
    end
    doMotion("X+Attack")
end

local function attack_dive(data)
    if data ~= nil then
        data.dir = dirs.U
        data.pressAction = false
    end
    param.attackCooldown = 20
    doMotion("U+Attack")
end

local function attack_headbutt(data)
    if data ~= nil then
        data.dir = dirs.D
        data.pressAction = false
    end
    doMotion("D+Attack")
end

local function attack_kidnap(data)
    if data ~= nil then
        data.dir = dirs.N
        data.pressAction = true
    end
    doMotion("X+Heruka")
end

local function turnToPlayer()
    local me, player = getMe(), getPlayer()
    local playerBehind = (player.XPos - me.XPos) * me.FacingDir
    if playerBehind < 0 then
      addInput(dirs.B)
    end
end

local _SPRITE_DETECTION_RANGE = 160 * 15
local function getSpritesOfMe()
    local me = getMe()
    local range = _SPRITE_DETECTION_RANGE / 2
    local l, r, u, d = me.XPos - range, me.XPos + range, me.YPos + range, me.YPos - range

    local sprites = getSpritesInBox(l, r, u, d, nil, me)
    local sprites_toReturn = {}
    for i, spr in ipairs(sprites) do
        if spr.MyName == me.MyName
        and not is_hurt(spr) 
        and not isTangled(spr) then
            sprites_toReturn[#sprites_toReturn+1] = spr
        end
    end

    return sprites_toReturn
end

local function getHurtSpritesOfMe()
  local me = getMe()
  local range = _SPRITE_DETECTION_RANGE / 2
  local l, r, u, d = me.XPos - range, me.XPos + range, me.YPos + range, me.YPos - range

  local sprites = getSpritesInBox(l, r, u, d, nil, me)
  local sprites_toReturn = {}
  for i, spr in ipairs(sprites) do
      if spr.MyName == me.MyName
      and (is_hurt(spr) 
      or isTangled(spr)) then
          sprites_toReturn[#sprites_toReturn+1] = spr
      end
  end

  return sprites_toReturn
end

-- the initialization function for this AI
function OnInit()
    -- Accessible anywhere in this file after Initialization
    local me, player = getMe(), getPlayer()
    --
    param.origin = {
        x = me.XPos,
        y = me.YPos
    }
    if me:getSpawnpointPos() ~= nil then
        param.origin.x, param.origin.y, param.origin.z = me:getSpawnpointPos()
    end
    --
    param.isLeader = false
    param.time = 0
    param.attackCooldown = 0 
    --
    param.flockBoundsX = 160 * 8
    param.flockBoundsY = { top = 160 * 8, bottom = 160 * -3}
end

-- the main update function for this AI
function OnUpdate()
  while true do
    local me, player = getMe(), getPlayer()
    local sprites = getSpritesOfMe()
    local hurtSprites = getHurtSpritesOfMe()
    if aiDebug() then
      draw_bar(param.attackCooldown, 0, 20, 0, {255, 0, 0})
    end
    
    if tick_ok()
    and not is_hurt() then
      param.attackCooldown = math.max(0, param.attackCooldown - (1))
    end

    local function BoundRanges()
      local bL, bR, bT, bB = 0
      bL = param.origin.x - param.flockBoundsX 
      bR = param.origin.x + param.flockBoundsX
      bT = param.origin.y + param.flockBoundsY.top
      bB = param.origin.y + param.flockBoundsY.bottom
      return bL, bR, bT, bB
    end

    local vecCenter = boids_mass(sprites)
    local vecBias = boids_bias(me, sprites, 300)
    local vecAvgVelocity = boids_vel(sprites)
    local vecBounds = boids_bounds(me, BoundRanges() ) 
    local vecTendency = boids_tendency(
      { x = param.origin.x, y = param.origin.y, z = param.origin.z},
      me
    )
    local vecRepulse = boids_tendency(
      { x = player.XPos, y = player.YPos, z = player.ZPos},
      me
    )

    local vecHurtAvoid = boids_mass(hurtSprites)

    local vecAvoidPoint = {x = 0, y = 0, z = 0}
    
    local numFriendsAttacking = 0 
    for i, spr in ipairs(sprites) do
        if spr:hasTag("Attack") then
            local pos = { x = spr.XPos, y = spr.YPos, z = spr.ZPos }
            vecAvoidPoint = boids_tendency(pos, me)
            numFriendsAttacking = numFriendsAttacking + 1
        end
    end

    vecAvoidPoint.x = vecAvoidPoint.x / 1
    vecAvoidPoint.y = vecAvoidPoint.y / 1
    vecAvoidPoint.z = vecAvoidPoint.z / 1

    vecCenter.x = (vecCenter.x - me.XPos) / 100
    vecCenter.y = (vecCenter.y - me.YPos) / 100
    vecCenter.z = (vecCenter.z - me.ZPos) / 100

    local xVelWS = me.XVel * me.FacingDir
    vecAvgVelocity.x = (vecAvgVelocity.x - xVelWS) / 8
    vecAvgVelocity.y = (vecAvgVelocity.y - me.yVel) / 8
    vecAvgVelocity.z = (vecAvgVelocity.z - me.zVel) / 8

    if not is_hurt(me) then
      local xDiff = vecBounds.x 
      + vecTendency.x / 80
      - vecAvoidPoint.x / 250
      -- xDiff = clamp(xDiff, -2, 2)
      
      local yDiff = vecBounds.y 
      + vecTendency.y / 80
      - vecAvoidPoint.y / 250

      if #sprites > 0 then
        xDiff = xDiff
        + vecCenter.x 
        + vecBias.x 
        + vecAvgVelocity.x
        yDiff = yDiff
        + vecCenter.y 
        + vecBias.y 
        + vecAvgVelocity.y
      end

      if #hurtSprites > 0 then
        xDiff = xDiff - ( vecHurtAvoid.x * 0.125 ) 
        yDiff = yDiff - ( vecHurtAvoid.y * 0.125 )
      end

      local playerVecDist = vec_dist({
          x = player.XPos,
          y = player.YPos
      },
      {
          x = me.XPos,
          y = me.YPos,
      })

      -- Use this to change our facing Dir
      local facingDestPoint = { 
        x = param.origin.x,
        y = param.origin.y
      }

      local anyoneAttacking = false
      local anyoneInBattle = false

      for i, spr in ipairs(sprites) do
        if spr:hasTag("Attack") then 
          anyoneAttacking = true
        end
        if spr.InBattle > 0 then
        --or spr.WasInBattle >= 0 then
          anyoneInBattle = true
        end
      end

      if playerVecDist <= 160 * 7 then
        facingDestPoint.x = player.XPos
        facingDestPoint.y = player.YPos
        --
        xDiff = xDiff - vecRepulse.x * 0.125 
        yDiff = yDiff - vecRepulse.y * 0.125
        -- Don't bother avoiding the player in battle, but by a bit
        -- and remove the bounds and bias
        if anyoneInBattle then
          xDiff = xDiff - (vecBounds.x/2) + (vecAvoidPoint.x / 275) - (vecBias.x / 3)
          yDiff = yDiff - (vecBounds.y/2) + (vecAvoidPoint.y / 275) - (vecBias.y / 3)
        end
      end

      -- Attack --
      if playerVecDist <= 160 * 2.5 
      and not anyoneAttacking 
      and player_vulnerable()
      and canAttack(me)
      and param.attackCooldown <= 0 then
        turnToPlayer()
        attack_dive()
      end

      local velocityDelta = { x = xDiff, y = yDiff }
      local deltaLen = vec_len(velocityDelta)
      
      local MAX_DELTA = 1.75
      if deltaLen > MAX_DELTA then
        xDiff = xDiff * (MAX_DELTA / deltaLen)
        yDiff = yDiff * (MAX_DELTA / deltaLen)
      end
      
      xVelWS = xVelWS + xDiff
      me.XVel = xVelWS * me.FacingDir
      me.YVel = me.YVel + yDiff

      local velocityVector = { x = me.xVel, y = me.YVel }
      local velocityLen = vec_len(velocityVector)
      
      local MAX_VELOCITY = 26
      if velocityLen > MAX_VELOCITY then
        me.XVel = me.XVel * (MAX_VELOCITY / velocityLen)
        me.YVel = me.YVel * (MAX_VELOCITY / velocityLen)
      end

      -- face the tendency!
      local shouldTurnAround = (facingDestPoint.x - me.XPos) * me.FacingDir
      if shouldTurnAround < 0 then
        addInput(dirs.B)
      end

      if aiDebug() then
        drawBox( me.XPos + vecRepulse.x - 80, me.XPos + vecRepulse.x + 80, me.YPos + vecRepulse.y - 80, me.YPos + vecRepulse.y + 80,
        1, -1, 255, 255, 255 )

        drawBox( me.XPos + xDiff - 80, me.XPos + xDiff + 80, me.YPos + yDiff - 80, me.YPos + yDiff + 80,
        1, -1, 0, 0, 255 )
      end

    -- set up useful variables on the sprite
      me.LightZ = AIFlyingHash
      me.Friction = 0
      me.Weight = 0
      me.PhysicsType = 0
    end
    wait() -- try to keep this at the end
  end -- the while loop
end -- the function

function OnBattleAttack(battleData)
    local toReturn = DefaultOnBattleAttack(battleData)
    local me = getMe()

    -- Cihuateteo
    attack_swipe(toReturn)

    return toReturn
end --OnBattleAttack()